/* file "process.cc" */

/*  Copyright (c) 1994 Stanford University

    All rights reserved.

    This software is provided under the terms described in
    the "suif_copyright.h" include file. */

#include <suif_copyright.h>

/* Create C++ files */

#include <stdlib.h>
#include <suif1/misc.h>
#include "lex.h"
#include "symbol.h"
#include <string.h>
#include <malloc.h>
#include <ctype.h>

#define LPAREN '('
#define RPAREN ')'
#define CPS(f) code_stream->put_string(f);
#define DPS(f) defs_stream->put_string(f);
#define Sk_top (-3)
#define Sk_bottom (-2)
#define Sk_identity (-1)

void mk_class_header(const char *,const char *);
void mk_class_header2(const char *,const char *,o_stream *);
void mk_class(const char *,const char *,C_code *);
char *mk_name(const char *,const char *);
extern void yyerror(const char *);

/*

Sharlit writes its output, the analyzer sources, into two
files, a .h file and a .c file.  To write to the .h file,
Sharlit uses defs_stream, and to write to the .c file, code_stream.

Experimental code within sharlit uses two alternative streams file,
nheader_stream and ncode_stream.  They correspond respectively to
defs_stream and code_stream.

*/
extern o_stream *defs_stream, *code_stream;
o_stream *nheader_stream, *ncode_stream;

ty_base *void_type, *int_type, *char_type;
ty_pointer *int_pointer_type;
extern Symbol_table *symtab;
Symbol *problem_name;
SV_dflow *problem;

/*##
These suffixes are appended to the basic problem identifier to
name C++ classes.  For example, the {\em per problem} structure
for a data flow problem, @df@ would be called @df_P@.
The @PP_suffix@, @SS_suffix@, @LL_suffix@, and @VV_suffix@ is for the
per problem class, the skeleton class, local transformer base class,
and the variable class, respectively.

*/
const char *PP_suffix = "_P";
const char *SS_suffix = "_S";
const char *LL_suffix = "_L";
const char *VV_suffix = "_V";

/*

Called to set the name of the analyzer being generated. Since Sharlit
generates one analyzer per input file, it only makes sense to call
set_name once, and set_name checks this condition.

*/
void
set_name(Symbol *n)
{
    assert(problem_name==0);
    TAG_ASSERT(SV_dflow,n);
    problem_name=n;
    problem=(SV_dflow *)n->value;
}

void
cleanup()
{
    problem=0;
    problem_name=0;
}

ty_record *sgraph, *snode, *strans, *svar, *seqn;
ty_named *snode_type, *svar_type, *strans_type, *seqn_type;
ty_pointer *snode_pointer, *strans_pointer, *svar_pointer, *seqn_pointer;
ty_record *sproblem, *sprob_iter;
ty_record *sprob_bb, *sprob_tarjan, *sprob_tarjan_debug;
ty_record *sproblem_debug, *sprob_switches, *sprob_switches_test;
ty_record *pp_stack;
ty_record *pp_stack_e;
ty_named *pp_stack_e_type;
ty_named *pp_stack_type;
ty_pointer *pp_stack_pointer;
ty_enum *cfgnode_kinds;
ty_named *cfgnode_kinds_type;

#define DEFINE_PROBLEM(x,y) \
	s=new Symbol(x); \
	s->value=y=new ty_record(s,0); \
	symtab->enter(s);
#define NEW_RECORD(x,y)   x=new ty_record(new Symbol(y),0)
#define NEW_NAMED(x,y)    x=new ty_named(y->type_name,y)
#define NEW_ENUM(x,y)     x=new ty_enum(new Symbol(y))

void
init_types()
{
    Symbol *s;
    void_type=new ty_base(BT_void);
    int_type=new ty_base(BT_int);
    char_type=new ty_base(BT_char);
    int_pointer_type=new ty_pointer(int_type);
    
    NEW_RECORD(sgraph,"Sgraph");
    NEW_RECORD(strans,"Strans");
    NEW_RECORD(svar,"Svar");
    NEW_RECORD(snode,"Snode");
    NEW_RECORD(seqn,"Seqn");
    NEW_RECORD(pp_stack,"PP_stack");
    NEW_RECORD(pp_stack_e,"PP_stack_e");
    
    NEW_NAMED(strans_type,strans);
    NEW_NAMED(svar_type,svar);
    NEW_NAMED(snode_type,snode);
    NEW_NAMED(seqn_type,seqn);
    NEW_NAMED(pp_stack_type,pp_stack);
    NEW_NAMED(pp_stack_e_type,pp_stack_e);
    
    svar_pointer=new ty_pointer(svar_type);
    strans_pointer=new ty_pointer(strans_type);
    snode_pointer=new ty_pointer(snode_type);
    seqn_pointer=new ty_pointer(seqn_type);
    pp_stack_pointer=new ty_pointer(pp_stack_type);
    
    NEW_ENUM(cfgnode_kinds,"CFGnode_kinds");
    NEW_NAMED(cfgnode_kinds_type,cfgnode_kinds);

    snode->add_member("stacked",char_type);
    snode->add_member("improper",char_type);
    snode->add_member("G",svar_pointer);
    snode->add_member("L",strans_pointer);
    
    strans->add_member("kind",int_type);
    
    DEFINE_PROBLEM("Sproblem",sproblem);
    DEFINE_PROBLEM("Sprob_iter",sprob_iter);
    DEFINE_PROBLEM("Sprob_bb",sprob_bb);
    DEFINE_PROBLEM("Sprob_tarjan",sprob_tarjan);
    DEFINE_PROBLEM("Sprob_tarjan_DEBUG",sprob_tarjan_debug);
    DEFINE_PROBLEM("Sproblem_DEBUG",sproblem_debug);
    DEFINE_PROBLEM("Sprob_switches",sprob_switches);
    DEFINE_PROBLEM("Sprob_switches_test",sprob_switches_test);
}

/*

Initialize the files generated by sharlit

*/
extern char *input_base, *defs_file, *define_name;
extern int generalized;

void init_files()
{
    int i;
    char *cp;
    if(strcmp(defs_file,"-")==0)
      define_name=0;
    else{
	cp=define_name=new char[strlen(input_base)+10];
	*cp++ = '_' ;
	for(i=0;input_base[i];i++){
	    if(islower(input_base[i]))
	      *cp =input_base[i] - 'a' + 'A';
	    else if(isupper(input_base[i])||isdigit(input_base[i]))
	      *cp =input_base[i];
	    else *cp ='_';
	    cp++;
	}
	*cp=0;
	DPS("#ifndef ");
	DPS(define_name);
	DPS("\n#define ");
	DPS(define_name);
	DPS("\n");
	DPS("#pragma interface\n");
	if(generalized){
	    nheader_stream->put_string("#ifndef ");
	    nheader_stream->put_string(define_name);
	    nheader_stream->put_string("\n#define ");
	    nheader_stream->put_string(define_name);
	    nheader_stream->put('\n');
	}
    }
}

void init_code_file(int flag)
{
    code_stream->put_string("#pragma implementation\n");
    code_stream->put_string("#include <nsharlit/sharlit.h>\n");
    if(flag)
      {
	code_stream->put_string("#include <stdio.h>\n");
	code_stream->put_string("#include <nsharlit/cfg.h>\n");
	code_stream->put_string("#include <nsharlit/sharlit_test.h>\n");
      }
}

void finish_files()
{
    if(!define_name)
      return;
    defs_stream->put_string("#endif\n");
    if(generalized)
      nheader_stream->put_string("#endif\n");
}

/*

Initialize the structure that represents a data-flow analyzer.  There
is an enumerated type generated but it isn't registered in the main
symbol table.  We may want to do this someday.

*/

SV_dflow *create_dflow(Symbol *n,ty_record *b)
{
    SV_dflow *df;
    if(b==sprob_switches){
	df=new SV_dflow_switches(n,b);
	df->new_version=1;
	return df;
    }else{
	return new SV_dflow(n,b);
    }
}

SV_dflow::SV_dflow(Symbol *n,ty_record *b):Sym_val(TAG(SV_dflow)),name(n)
{
    char *enum_name=mk_name(n->name,"_kinds");
    n_nodes=0;
    new_version=0;
    SS=0;
    LL=0;
    VV=0;
    unique=0;
    VV_pointer_type=0;
    new_var_body=0;
    copy_var_body=0;
    meet_body=0;
    kinds=new ty_enum(new Symbol(enum_name));
    delete enum_name;
    solver_type=new ty_record(n,b);
    solver_pointer_type=new ty_pointer(new ty_named(n,solver_type));
    flow_map=0;
    identity_function=0;
}

SV_dflow::~SV_dflow()
{
    if (kinds != NULL)
	delete kinds;
    if (solver_type != NULL)
	delete solver_type;
    if (SS != NULL)
	delete SS;
    if (LL != NULL)
	delete LL;
    if (VV != NULL)
	delete VV;
    if (solver_pointer_type != NULL)
	delete solver_pointer_type;
}

SV_dflow_parsing::SV_dflow_parsing(Symbol *n,ty_record *b)
     :SV_dflow(n,b)
{
}
SV_dflow_switches::SV_dflow_switches(Symbol *n,ty_record *b)
     :SV_dflow_parsing(n,b)
{
}


/*##

This routine writes out the class definitions for the
data flow information.

*/
void SV_dflow::define_node(Symbol *sp)
{
    char *enum_member=mk_name("Sk_",sp->name);
    SV_node *nv=(SV_node *)sp->value;
    kinds->add_number(enum_member);
    delete enum_member;
    nodes.append(sp);
    nv->index=n_nodes++;
    nv->df=this;
}

void SV_dflow::define_identity(Symbol *sp)
{
    SV_node *nv=(SV_node *)sp->value;
    define_node(sp);
    nv->is_identity=1;
    assert(identity_function==0);
    identity_function=sp;
}

/*##
frag(define_PSV,"Define @PP@, @SS@, and @VV@")

*/
void
SV_dflow::create_solver_type(C_code *def)
{
    assert(solver_type!=0);
    assert(solver_type->opaque==0);
    solver_type->opaque=def;
}
void
SV_dflow::define_SS(C_code *def,ty_record *b)
{
	Symbol *ss;
	char *ss_name=mk_name(name->name,SS_suffix);
	assert(SS==0);
	SS=new ty_record(ss=new Symbol(ss_name),b);
	SS_pointer_type=new ty_pointer(new ty_named(ss,SS));
	SS->opaque=def;
	delete ss_name;
}
void
SV_dflow::define_VV(Symbol *n,C_code *def)
{
	char *gname=0;
	if(!n){
		gname=mk_name(name->name,VV_suffix);
		n=new Symbol(gname);
	}
	assert(VV==0);
	VV=new ty_record(n,svar);
	VV->opaque=def;
	VV_pointer_type=new ty_pointer(new ty_named(VV->type_name,VV));
	delete gname;
}
void
SV_dflow::define_LL(C_code *def)
{
    char *lname=mk_name(name->name,LL_suffix);
    assert(LL==0);
    LL=new ty_record(new Symbol(lname),strans);
    LL->add_member("P",solver_pointer_type);
    if(!new_version)
      LL->add_member("S",SS_pointer_type);
    LL->opaque=def;
    delete lname;
}

/*##
frag(generate,"Generate the node classes and transformers")

*/
void
SV_dflow::code_gen(CG_context cntx, o_stream *s)
{
    Symbol_iter si;
    switch(cntx){
      case CGC_header:
	kinds->code_gen(cntx,s);
	s->put_string(";\n");
	break;
      default:
	/* maybe should be fall through, but just in case do the assert */
	assert(0);
    }
    solver_type->code_gen(cntx,s);
    SS->code_gen(cntx,s);
    VV->code_gen(cntx,s);
    for(si.reset(&extra_classes);!si.is_empty();)
      si.step()->value->code_gen(cntx,s);
    LL->code_gen(cntx,s);
    for(si.reset(&nodes);!si.is_empty();)
      si.step()->value->code_gen(cntx,s);
}


extern int debug_flag;
void
SV_dflow::generate()
{
    Symbol *s;
    Symbol_iter si;

    /* declare the server object */
    declare_kinds();
    declare_generated_functions();
    solver_type->declare(defs_stream,code_stream,0);

    if(SS==0)
      define_SS(0,snode);
    SS->declare(defs_stream,code_stream,0);
    VV->declare(defs_stream,code_stream,0);
    for(si.reset(&extra_classes);!si.is_empty();)
      si.step()->value->declare(defs_stream,code_stream,0);
    if(LL==0)
      define_LL(0);
    LL->declare(defs_stream,code_stream,0);
    for(si.reset(&nodes);!si.is_empty();){
	s=si.step();
	((SV_node *)s->value)->gen(s,this);
    }
    for(si.reset(&nodes);!si.is_empty();){
	s=si.step();
	((SV_node *)s->value)->gen_func(s,this);
    }
    if(new_var_body){
	CPS("Svar *");
	CPS(solver_type->type_name->name);
	CPS("::new_var(){\n");
	code_stream->put_code(new_var_body);
	CPS("}\n");
    }
    if(copy_var_body){
	CPS("int ");
	CPS(solver_type->type_name->name);
	CPS("::copy_var(Svar *__d,Svar *__s){\n");
	VV_pointer_type->print(code_stream);
	CPS(" _DST = ");
	code_stream->put(LPAREN);
	VV_pointer_type->print(code_stream);
	code_stream->put(RPAREN);
	CPS(" __d;\n");
	VV_pointer_type->print(code_stream);
	CPS(" _SRC = ");
	code_stream->put(LPAREN);
	VV_pointer_type->print(code_stream);
	code_stream->put(RPAREN);
	CPS(" __s;\n");
	code_stream->put_code(copy_var_body);
	CPS("}\n");
    }
    if(meet_body){
	CPS("void ");
	CPS(solver_type->type_name->name);
	CPS("::meet(Svar *__d,Snode *__n){\n");
	VV_pointer_type->print(code_stream);
	CPS(" _DST = ");
	code_stream->put(LPAREN);
	VV_pointer_type->print(code_stream);
	code_stream->put(RPAREN);
	CPS(" __d;\n");
	VV_pointer_type->print(code_stream);
	CPS("*_SRCS = 0;\n");
	CPS("int _NSRCS=get_srcs(*(Svar ***)&_SRCS,__n);\n");
	code_stream->put_code(meet_body);
	CPS("}\n");
    }
    generate_flow_map();
    generate_simplifier();
}

void
SV_dflow::declare_kinds()
{
    kinds->print(defs_stream);
    defs_stream->put_string(";\n");
}

void
SV_dflow::declare_generated_functions()
{
}

void
SV_dflow::generate_simplifier()
{
}

void
SV_dflow::generate_flow_map()
{
    assert(flow_map==0);
}


/* build star rule */
void
SV_dflow_parsing::generate_STAR()
{
    Symbol_iter si;
    Symbol *s;
    SV_path *p;
    SV_rule_iter ri;
    SV_rule *r;

    CPS("Strans *");
    CPS(solver_type->type_name->name);
    CPS("::STAR(Strans *_T,Seqn *_Q){\n");
    CPS("switch(_T->kind){\n");
    for(si.reset(&nodes);!si.is_empty();){
	s=si.step();
	if(TAG_IS(SV_path,s)){
	    p=(SV_path *)s->value;
	    if(p->rules[TAG(SVR_star)].is_empty())
	      continue;
	    CPS("case Sk_");
	    CPS(s->name);
	    CPS(":\n");
	    for(ri.reset(&p->rules[TAG(SVR_star)]);!ri.is_empty();){
		r=ri.step();
		TAG_ASSERT(SVR_star,r);
		r->generate(p);
	    }
	}
    }
    CPS("case Sk_top: return &strans_identity;\n");
    CPS("case Sk_bottom: break;\n");
    CPS("default: delete _T;\n");
    CPS("}\nreturn &strans_bottom;\n}\n");
}

/* build meet rule */
void
SV_dflow_parsing::generate_MEET()
{
    Symbol_iter si;
    Symbol *s;
    SV_path *p;
    SV_rule_iter ri;
    SV_rule *r;
    CPS("Strans *");
    CPS(solver_type->type_name->name);
    CPS("::MEET(Strans *_T,Strans *_R,Seqn *_Q){\n");
    CPS("switch(_T->kind){\n");
    for(si.reset(&nodes);!si.is_empty();){
	s=si.step();
	if(TAG_IS(SV_path,s)){
	    p=(SV_path *)s->value;
	    if(p->rules[TAG(SVR_meet)].is_empty())
	      continue;
	    CPS("case Sk_");
	    CPS(s->name);
	    CPS(":{switch(_R->kind){\n");
	    for(ri.reset(&p->rules[TAG(SVR_meet)]);!ri.is_empty();){
		r=ri.step();
		TAG_ASSERT(SVR_meet,r);
		r->generate(p);
	    }
	    CPS("}\nbreak;\n}\n");
	}
    }
    CPS("}\nreturn MEET_uu(_T,_R);\n}\n");
}

/* build NONTERMINAL function */
void
SV_dflow_parsing::generate_NONTERMINAL()
{
    Symbol_iter si;
    Symbol *s;
    SV_node *n;
    CPS("int ");
    CPS(solver_type->type_name->name);
    CPS("::NONTERMINAL(Strans *l){\n");
    CPS("switch(l->kind){\n");
    for(si.reset(&nodes);!si.is_empty();){
	s=si.step();
	n=(SV_node *)s->value;
	if(n->nonterminal){
	    CPS("case Sk_");
	    CPS(s->name);
	    CPS(": return 1;\n");
	}
    }
    CPS("case Sk_top: case Sk_identity: case Sk_bottom:");
    CPS("return 1;\n");
    CPS("}\nreturn 0;\n}\n");
}

void
SV_dflow_parsing::declare_generated_functions()
{
    ty_function *ft;
    assert(new_version);

#define NEW_MEMBER_FUNC(f,t,name) \
    f=new ty_function(t,solver_type); \
    solver_type->add_member(name,f);

    if(new_var_body){
	NEW_MEMBER_FUNC(ft,svar_pointer,"new_var");
    }
    if(copy_var_body){
	NEW_MEMBER_FUNC(ft,int_type,"copy_var");
	ft->add_formal(svar_pointer);
	ft->add_formal(svar_pointer);
    }
    if(meet_body){
	NEW_MEMBER_FUNC(ft,void_type,"meet");
	ft->add_formal(svar_pointer);
	ft->add_formal(snode_pointer);
    }
    NEW_MEMBER_FUNC(ft,void_type,"_pp_init");
    NEW_MEMBER_FUNC(ft,strans_pointer,"STAR");
    ft->add_formal(strans_pointer);
    ft->add_formal(seqn_pointer);
    NEW_MEMBER_FUNC(ft,strans_pointer,"MEET");
    ft->add_formal(strans_pointer);
    ft->add_formal(strans_pointer);
    ft->add_formal(seqn_pointer);
    NEW_MEMBER_FUNC(ft,pp_stack_e_type,"_pp_reduce");
    ft->add_formal(strans_pointer);
    ft->add_formal(seqn_pointer);
    ft->add_formal(pp_stack_pointer);
    NEW_MEMBER_FUNC(ft,int_type,"NONTERMINAL");
    ft->add_formal(strans_pointer);
#undef NEW_MEMBER_FUNC
}
void
SV_dflow_parsing::generate_simplifier()
{
    PP_mach pm(n_nodes);
    int i;
    PP_state *s=new PP_state(n_nodes);
    Symbol_list *sl;
    PP_state *ns;
    Symbol_iter si;
    assert(new_version);
    s->add(&rules);
    pm.add(s);
    for(i=0;i<pm.n_states;i++){
	s=pm.states[i];
	assert(s!=0);
	s->next_set(sl=new Symbol_list);
	for(si.reset(sl);!si.is_empty();){
	    Symbol *sp=si.step();
	    int has_reductions;
	    int index=sp->node_value()->index;
	    ns=s->apply(sp,&has_reductions);
	    if(ns)
	      s->out_edges[index]=pm.add(ns);
	    else if(has_reductions)
	      s->out_edges[index]=-1;/* do_reduction */
	    else
	      s->out_edges[index]=-2;/* stuck state */
	}
	if(debug_flag){
	    s->print(stderr);
	    sl->print(stderr);
	    putc('\n',stderr);
	}
	delete sl;
    }
    pm.generate(code_stream,this);
    generate_STAR();
    generate_MEET();
    generate_NONTERMINAL();
}
void
SV_dflow_switches::declare_generated_functions()
{
    ty_function *ft;
    assert(new_version);
    if(new_var_body){
	ft=new ty_function(svar_pointer,solver_type);
	solver_type->add_member("new_var",ft);
    }
    if(copy_var_body){
	ft=new ty_function(int_type,solver_type);
	ft->add_formal(svar_pointer);
	ft->add_formal(svar_pointer);
	solver_type->add_member("copy_var",ft);
    }
    if(meet_body){
	ft=new ty_function(void_type,solver_type);
	ft->add_formal(svar_pointer);
	ft->add_formal(snode_pointer);
	solver_type->add_member("meet",ft);
    }
#define NEW_MEMBER_FUNC(f,t,name) \
    f=new ty_function(t,solver_type); \
      solver_type->add_member(name,f);
    
    NEW_MEMBER_FUNC(ft,void_type,"_init_tables");
    if(has_rules()){
	NEW_MEMBER_FUNC(ft,strans_pointer,"_TIMES_xs");
	ft->add_formal(int_type);
	ft->add_formal(strans_pointer);
	ft->add_formal(snode_pointer);
	ft->add_formal(strans_pointer);
	ft->add_formal(snode_pointer);
	
	NEW_MEMBER_FUNC(ft,strans_pointer,"_ABSORB");
	ft->add_formal(strans_pointer);
	ft->add_formal(strans_pointer);
	ft->add_formal(snode_pointer);
	
	NEW_MEMBER_FUNC(ft,strans_pointer,"_CREATE");
	ft->add_formal(int_type);
	
	NEW_MEMBER_FUNC(ft,strans_pointer,"_UNIT");
	ft->add_formal(int_type);
	ft->add_formal(strans_pointer);
	ft->add_formal(snode_pointer);
	
	NEW_MEMBER_FUNC(ft,strans_pointer,"_MEET_ux");
	ft->add_formal(strans_pointer);
	ft->add_formal(strans_pointer);
	ft->add_formal(snode_pointer);
	ft->add_formal(int_type);
	
	NEW_MEMBER_FUNC(ft,strans_pointer,"_STAR_u");
	ft->add_formal(strans_pointer);
	ft->add_formal(snode_pointer);
	
	NEW_MEMBER_FUNC(ft,int_type,"NONTERMINAL");
	ft->add_formal(strans_pointer);

    }
    NEW_MEMBER_FUNC(ft,strans_pointer,"_FLOW_MAP");
    ft->add_formal(snode_pointer);
    ft->add_formal(cfgnode_kinds_type);
#undef NEW_MEMBER_FUNC
}

#define MEET_PREAMBLE "\n\
if(!_m){\n\
    if(!NONTERMINAL(_f))\n\
      return _MEET_ux(_TIMES_xs(1,_f,0,&strans_identity,0),\n\
		      _g,_gn,_no_remove);\n\
    if(!NONTERMINAL(_g)){\n\
	_f = _MEET_ux(_f,_TIMES_xs(1,_g,_gn,&strans_identity,0),_gn,0);\n\
	if(!_no_remove)\n\
	  delete _g;\n\
	return _f;\n\
    }\n\
    delete _f;\n\
    if(!_no_remove)\n\
      delete _g;\n\
    return &strans_bottom;\n\
}\n\
if(_m<0)\n\
     return _no_remove ? &strans_bottom : _MEET_ux(_g,_f,0,0);\n\
"

void SV_dflow_switches::generate_simplifier()
{
    char buf[1024];
    Symbol_iter si,fi,gi;
    SV_rule_list *rl;
    SV_rule_iter ri;
    Symbol *n,*s;
    SV_node *f,*g;
    Symbol_list unit_symbols,absorb_symbols;
    Symbol_list first,second;
    SV_path *p;
    Symbol **unit_table, **create_table, **absorb_table;
    int has_identity = (identity_function!=0);
    int *table,*ip,*jp,i,j,*copiable,*meet_able;
    int offset=Sk_top;
    int limit = n_nodes-Sk_top;
    
    /*
      tables summarizing the unit, create, absorb
	*/
    unit_table=new Symbol *[limit];
    create_table=new Symbol *[limit];
    absorb_table=new Symbol *[limit];
    table=new int[limit*limit];
    copiable=new int[limit];
    meet_able=new int[limit*limit];
    for(ip=table,jp=meet_able,i=0;i<limit;i++){
	unit_table[i]=create_table[i]=absorb_table[i]=0;
	copiable[i]=0;
	for(j=0;j<limit;j++){
	    *ip++=0;
	    *jp++=0;
	}
    }
    for(si.reset(&nodes);!si.is_empty();){
	n=si.step();
	if(!TAG_IS(SV_path,n))
	  continue;
	p=TAG_CHECK(SV_path,n);
	p->unit_table=new char[limit];
	p->create_table=new char[limit];
	p->absorb_table=new char[limit];
	for(i=0;i<limit;i++)
	  p->unit_table[i]=p->create_table[i]=p->absorb_table[i]=0;
	
	/*
	  extract all the symbols that are convertible to the path
	    using a single unit rule
	      */
	ri.reset(&p->rules[TAG(SVR_unit)]);
	while(!ri.is_empty()){
	    s=TAG_CHECK(SVR_unit,ri.step())->src_type;
	    if(TAG_IS(SV_node,s)||TAG_IS(SV_path,s))
	      f=(SV_node *)s->value;
	    unit_symbols.append(s);
	    unit_table[f->index-offset]=n;
	    p->unit_table[f->index-offset]=1;
	}
	
	/*
	  extract all the symbols that can be absorbed to the path
	    */
	ri.reset(&p->rules[TAG(SVR_absorb)]);
	while(!ri.is_empty()){
	    s=TAG_CHECK(SVR_absorb,ri.step())->src_type;
	    if(TAG_IS(SV_node,s)||TAG_IS(SV_path,s))
	      f=(SV_node *)s->value;
	    absorb_symbols.append(s);
	    absorb_table[f->index-offset]=n;
	    p->absorb_table[f->index-offset]=1;
	}
	rl = &p->rules[TAG(SVR_create)];
	if(!rl->is_empty()){
	    /*
	      A create rule is present, so the first symbols is
		a union of unit_symbols and absorb_symbols
		  */
	    if(unit_symbols.non_empty_intersect(absorb_symbols)){
		yyerror("ambiguous rules");
		exit(1);
	    }
	    for(fi.reset(&absorb_symbols);!fi.is_empty();){
		s=fi.step();
		if(TAG_IS(SV_node,s)||TAG_IS(SV_path,s))
		  f=(SV_node *)s->value;
		create_table[f->index-offset]=n;
		p->create_table[f->index-offset]=1;
	    }
	    first += unit_symbols;
	    first += absorb_symbols;
	    second += absorb_symbols;
	} else {
	    /*
	      A create rule is not present, so the first symbol is
		the unit_symbols, and the second symbol is the
		  absorb_symbols.
		    */
	    first += unit_symbols;
	    second += absorb_symbols;
	}
	ri.reset(&p->rules[TAG(SVR_meet)]);
	while(!ri.is_empty()){
	    SVR_meet *meet_rule=TAG_CHECK(SVR_meet,ri.step());
	    assert(meet_rule->lhs==n);
	    s=meet_rule->rhs;
	    if(TAG_IS(SV_node,s)||TAG_IS(SV_path,s))
	      f=(SV_node *)s->value;
	    ip=&meet_able[(p->index-offset)*limit+f->index-offset];
	    jp=&meet_able[(f->index-offset)*limit+p->index-offset];
	    if(*ip || *jp){
		yyerror("ambiguous meet");
	    }
	    *ip=1;
	    if(n!=s)
	      *jp=-1;
	}
	if(debug_flag){
	    unit_symbols.print(stderr);
	    putc('\n',stderr);
	    absorb_symbols.print(stderr);
	    putc('\n',stderr);
	    first.print(stderr);
	    putc('\n',stderr);
	    second.print(stderr);
	    putc('\n',stderr);
	}
	for(fi.reset(&first);!fi.is_empty();){
	    s=fi.step();
	    if(TAG_IS(SV_node,s)||TAG_IS(SV_path,s))
	      f=(SV_node *)s->value;
	    ip=&copiable[f->index-offset];
	    if(*ip){
		yyerror("ambiguous copy");
		exit(1);
	    }
	    *ip=1;
	    if(f->is_identity)
	      copiable[Sk_identity-offset]=1;
	}
	if(!p->rules[TAG(SVR_absorb)].is_empty())
	  if(!first.contains(n))
	    first.append(n);
	for(fi.reset(&first);!fi.is_empty();){
	    s=fi.step();
	    if(TAG_IS(SV_node,s)||TAG_IS(SV_path,s))
	      f=(SV_node *)s->value;
	    for(gi.reset(&second);!gi.is_empty();){
		s=gi.step();
		if(TAG_IS(SV_node,s)||TAG_IS(SV_path,s))
		  g=(SV_node *)s->value;
		ip=&table[(f->index-offset)*limit+g->index-offset];
		if(*ip){
		    yyerror("ambiguous times");
		    exit(1);
		}
		*ip=1;
		if(f->is_identity)
		  table[(Sk_identity-offset)*limit+g->index-offset]=1;
	    }
	}
	if(debug_flag){
	    for(ip=table,i=0;i<limit;i++){
		for(j=0;j<limit;j++){
		    fprintf(stderr,"%2d",*ip++);
		}
		putc('\n',stderr);
	    }
	}
    } /* for(si.reset....) */
    if(has_rules()){
	    /* fixup table for sk_identity */
	for(i=Sk_identity-offset;i<limit;i++)
	  table[i*limit+Sk_identity-offset]=1;
	    /* fixup meet_able for sk_top */
	for(i=0;i<limit;i++){
	    meet_able[(Sk_top-offset)*limit+i]=1;
	    if(i!=Sk_top-offset)
	      meet_able[i*limit+Sk_top-offset]= -1;
	    meet_able[(Sk_bottom-offset)*limit+i]=1;
	    if(i!=Sk_bottom-offset)
	      meet_able[i*limit+Sk_bottom-offset]= -1;
	}
	CPS("static Sharlit_flag _copiable[]={\n");
	for(ip=copiable,i=0;i<limit;i++){
	    sprintf(buf,"%2d,",*ip++);
	    CPS(buf);
	}
	CPS("};\n");
	CPS("static Sharlit_state _times_able[]={\n");
	for(ip=table,i=0;i<limit;i++){
	    for(j=0;j<limit;j++){
		sprintf(buf,"%2d,",*ip++);
		CPS(buf);
	    }
	    CPS("\n");
	}
	CPS("};\n");
	CPS("static Sharlit_state _meet_able[]={\n");
	for(ip=meet_able,i=0;i<limit;i++){
	    for(j=0;j<limit;j++){
		sprintf(buf,"%2d,",*ip++);
		CPS(buf);
	    }
	    CPS("\n");
	}
	CPS("};\n");
    }
    delete table;
    delete copiable;
    delete meet_able;

    /* function that initialize the simplifier's tables */
    sprintf(buf,"void %s::_init_tables()\n{\n",solver_type->type_name->name);
    CPS(buf);
    sprintf(buf,"assert(%d==Sk_top); /* check */\n",Sk_top);
    CPS(buf);
    sprintf(buf,"assert(%d==Sk_identity); /* check */\n",Sk_identity);
    CPS(buf);
    sprintf(buf,"n_states=%d;\n",has_rules() ? 10000 : 0);
    CPS(buf);
    if(has_rules()){
	CPS("_copiable=::_copiable;\n");
	CPS("_times_able=::_times_able;\n");
	CPS("_meet_able=::_meet_able;\n");
    }
    CPS("}\n");
    
    /* If there are no rules....do NOT do the rest of this stuff */
    if(!has_rules()){
	return;
    }
    if(debug_flag){
	fprintf(stderr,"unit: ");
	for(i=0;i<limit;i++)
	  fprintf(stderr,"%s, ",unit_table[i] ? unit_table[i]->name : "-");
	putc('\n',stderr);
	fprintf(stderr,"create: ");
	for(i=0;i<n_nodes;i++)
	  fprintf(stderr,"%s, ",create_table[i] ? create_table[i]->name : "-");
	putc('\n',stderr);
	fprintf(stderr,"absorb: ");
	for(i=0;i<n_nodes;i++)
	  fprintf(stderr,"%s, ",absorb_table[i] ? absorb_table[i]->name : "-");
	putc('\n',stderr);
    }
    
    /*
      Generate the times function
	*/
    sprintf(buf,"Strans *%s::_TIMES_xs",solver_type->type_name->name);
    CPS(buf);
    CPS("(int _needs_copy,Strans *_f,Snode *_fn,Strans *_g,Snode *_gn)\n");
    CPS("{\nStrans *_path;\n");
    if(!has_identity){
	CPS("if(_f==&strans_identity)\n");
	CPS("\treturn _TIMES_xs(1,_g,_gn,&strans_identity,0);\n");
    }
    sprintf(buf,"_copiable[_f->kind-(%d)]",offset);
    CPS("if(_needs_copy){ if(!");
    CPS(buf);
    CPS(") return &strans_bottom; }\n");
    sprintf(buf,"if(!_times_able[(_f->kind-(%d))*%d+_g->kind-(%d)]){\n",
	    offset,limit,offset);
    CPS(buf);
    CPS("\treturn &strans_bottom;\n}\n");
    /*
      The first symbol
	*/
    CPS("switch(_f->kind){\n");
    for(si.reset(&nodes);!si.is_empty();){
	s=si.step();
	if(TAG_IS(SV_node,s)||TAG_IS(SV_path,s))
	  f=(SV_node *)s->value;
	if(f->is_identity)
	  CPS("case Sk_identity:\n");
	sprintf(buf,"case Sk_%s:\n",s->name);
	CPS(buf);
	if(TAG_IS(SV_path,s)){
	    CPS("\n\tif(!_needs_copy) return _ABSORB(_f,_g,_gn);\n");
	}
	if(unit_table[f->index-offset]){
	    sprintf(buf,"convert %s to %s\n",s->name,
		    unit_table[f->index-offset]->name);
	    CPS(buf);
	}else if(create_table[f->index-offset]){
	    sprintf(buf,"\t_path=_CREATE(Sk_%s);\n",
		    create_table[f->index-offset]->name);
	    CPS(buf);
	    CPS("\t_path=_ABSORB(_path,_f,_fn);\n\tbreak;\n");
	}else
	  CPS("\tassert(0); /*stuck*/\n");
    }
    CPS("default:\n\tassert(0);\n}\n");
    
    /*
      The second symbol
	*/
    CPS("if(_g==&strans_identity) return _path;\n");
    CPS("\nreturn _ABSORB(_path,_g,_gn);\n}\n");
    delete unit_table;
    delete create_table;
    delete absorb_table;
    
    /* generate the create function */
    sprintf(buf,"Strans *%s::_CREATE(int _kind)\n",
	    solver_type->type_name->name);
    strcat(buf,"{\n\tStrans *_path;\n\tswitch(_kind){\n");
    CPS(buf);
    for(si.reset(&nodes);!si.is_empty();){
	n=si.step();
	p=TAG_IS(SV_path,n);
	if(!p)
	  continue;
	ri.reset(&p->rules[TAG(SVR_create)]);
	while(!ri.is_empty()){
	    SVR_create *create_rule=TAG_CHECK(SVR_create,ri.step());
	    sprintf(buf,"\tcase Sk_%s:\n{",create_rule->target_type->name);
	    CPS(buf);
	    create_rule->target_names->
	      bind_and_initialize2(code_stream,this,
				   create_rule->target_type,0,"_path");
	    if(create_rule->creator)
	      code_stream->put_code(create_rule->creator);
	    CPS("\n}\n\tbreak;\n");
	}
    }
    CPS("\n\tdefault: assert(0);\n\t}\n\treturn _path;\n}\n");
    
    /* generate the unit function */
    sprintf(buf,"Strans *%s::_UNIT(int _kind,Strans *_g,Snode *_gn)\n",
	    solver_type->type_name->name);
    strcat(buf,"{\n\tStrans *_path;\n\tswitch(_kind){\n");
    CPS(buf);
    for(si.reset(&nodes);!si.is_empty();){
	n=si.step();
	p=TAG_IS(SV_path,n);
	if(!p)
	  continue;
	ri.reset(&p->rules[TAG(SVR_unit)]);
	while(!ri.is_empty()){
	    SVR_unit *unit_rule=TAG_CHECK(SVR_unit,ri.step());
	    sprintf(buf,"\tcase Sk_%s:\n{",unit_rule->target_type->name);
	    CPS(buf);
	    unit_rule->target_names->
	      bind_and_initialize2(code_stream,this,
				   unit_rule->target_type,0,"_path");
	    unit_rule->qnames->
	      bind_and_initialize3(code_stream,this,
				   unit_rule->src_type,"_gn","_g");
	    if(unit_rule->creator)
	      code_stream->put_code(unit_rule->creator);
	    CPS("\n}\n");
	}
    }
    CPS("\n\tdefault: assert(0);\n\t}\n\treturn _path;\n}\n");
    
    /* generate the absorb function */
    sprintf(buf,"Strans *%s::_ABSORB(Strans *_f,Strans *_g,Snode *_gn)\n{\n",
	    solver_type->type_name->name);
    CPS(buf);
    CPS("if(_g==&strans_identity) return _f;\n");
    CPS("switch(_f->kind){\n");
    for(si.reset(&nodes);!si.is_empty();){
	n=si.step();
	p=TAG_IS(SV_path,n);
	if(!p)
	  continue;
	sprintf(buf,"\tcase Sk_%s:\n{\n",n->name);
	CPS(buf);
	CPS("\t\tswitch(_g->kind){\n");
	ri.reset(&p->rules[TAG(SVR_absorb)]);
	while(!ri.is_empty()){
	    SVR_absorb *absorb_rule=TAG_CHECK(SVR_absorb,ri.step());
	    sprintf(buf,"\tcase Sk_%s:\n\t\t{",absorb_rule->src_type->name);
	    CPS(buf);
	    absorb_rule->target_names->
	      bind_and_initialize3(code_stream,this,
				   absorb_rule->target_type,0,"_f");
	    absorb_rule->qnames->
	      bind_and_initialize3(code_stream,this,
				   absorb_rule->src_type,"_gn","_g");
	    if(absorb_rule->creator)
	      code_stream->put_code(absorb_rule->creator);
	    CPS("\n\t\t} break;\n");
	}
	CPS("\t\tdefault: assert(0);\n\t};break;}");
    }
    CPS("\n\tdefault: assert(0);\n\t}\n\treturn _f;\n}\n");
    
    /* generate meet function */
    sprintf(buf,
	    "Strans *%s::_MEET_ux(Strans *_f,Strans *_g,Snode *_gn,int _no_remove)\n",
	    solver_type->type_name->name);
    CPS(buf);
    sprintf(buf,"{\n\tint _m=_meet_able[(_f->kind-(%d))*%d+_g->kind-(%d)];\n",
	    offset,limit,offset);
    CPS(buf);
    CPS(MEET_PREAMBLE);
    CPS("switch(_f->kind){\n");
    CPS("case Sk_top: return _g;\n");
    CPS("case Sk_bottom: return _f;\n");
    for(si.reset(&nodes);!si.is_empty();){
	n=si.step();
	p=TAG_IS(SV_path,n);
	if(!p)
	  continue;
	sprintf(buf,"case Sk_%s:\n{\n",n->name);
	CPS(buf);
	CPS("\tswitch(_g->kind){\n");
	ri.reset(&p->rules[TAG(SVR_meet)]);
	while(!ri.is_empty()){
	    SVR_meet *meet_rule=TAG_CHECK(SVR_meet,ri.step());
	    sprintf(buf,"\tcase Sk_%s:\n\t{\n",meet_rule->rhs->name);
	    CPS(buf);
	    meet_rule->lhs_binder->
	      bind_and_initialize3(code_stream,this,
				   meet_rule->lhs,"_fn","_f");
	    meet_rule->rhs_binder->
	      bind_and_initialize3(code_stream,this,
				   meet_rule->rhs,"_gn","_g");
	    if(meet_rule->creator)
	      code_stream->put_code(meet_rule->creator);
	    CPS("\t}\n\tbreak;\n");
	}
	CPS("\tdefault: assert(0);\n\t}\nbreak;\n");
    }
    CPS("}\ndefault: assert(0);\n");
    CPS("\t}\nif(!_no_remove) delete _g;\nreturn _f;\n}\n");
    
    /* generate star function */
    sprintf(buf,"Strans *%s::_STAR_u(Strans *_f,Snode *_fn)\n",
	    solver_type->type_name->name);
    CPS(buf);
    CPS("{\nswitch(_f->kind){\n");
    CPS("case Sk_top: return &strans_identity;\n");
    CPS("case Sk_bottom: return &strans_bottom;\n");
    for(si.reset(&nodes);!si.is_empty();){
	SVR_star *star_rule;
	n=si.step();
	p=TAG_IS(SV_path,n);
	if(!p)
	  continue;
	ri.reset(&p->rules[TAG(SVR_star)]);
	if(ri.is_empty())
	  continue;
	sprintf(buf,"case Sk_%s:\n\t{\n",n->name);
	CPS(buf);
	star_rule=TAG_CHECK(SVR_star,ri.step());
	star_rule->target_binder->
	  bind_and_initialize3(code_stream,this,
			       star_rule->target_type,"_fn","_f");
	if(star_rule->creator)
	  code_stream->put_code(star_rule->creator);
	assert(ri.is_empty());
	CPS("\t}\n\tbreak;\n");
    }
    CPS("default:\n\tdelete _f;\n\t_f = &strans_bottom;\n");
    CPS("\t}\nreturn _f;\n}\n");
    
    generate_NONTERMINAL();
}

void SV_dflow_switches::generate_flow_map()
{
    char buf[1024];
    if(flow_map){
	sprintf(buf,"Strans *%s::_FLOW_MAP(Snode *__N, CFGnode_kinds _Kind)"
		"\n{\n\t",
		solver_type->type_name->name);
	CPS(buf);
	SS_pointer_type->print(code_stream);
	CPS(" _N = (");
	SS_pointer_type->print(code_stream);
	CPS(") __N;\n");
	code_stream->put_code(flow_map);
	CPS("}\n");
    }
}

void bind_and_initialize(Symbol *t,Symbol *s,const char *e)
{
    o_stream *cs=code_stream;
    cs->put_string(t->name);
    cs->put_string(" *");
    cs->put_string(s->name);
    cs->put_string("=(");
    cs->put_string(t->name);
    cs->put_string(" *)");
    cs->put_string(e);
    cs->put_string(";\n");
}

void SVR_star::generate(SV_node *n)
{
    o_stream *cs=code_stream;
    assert(target_node==n);
    cs->put_string("{\n");
    bind_and_initialize(target_type,target_binder->L_name,"_T");
    cs->put_code(creator);
    cs->put_string("} return _T;\n");
    return;
}
void
SVR_meet::generate(SV_node *n)
{
    o_stream *cs=code_stream;
    assert(lhs_node==n);
    if(rhs){
	cs->put_string("case Sk_");
	cs->put_string(rhs->name);
	cs->put_string(":{\n");
	bind_and_initialize(lhs,lhs_binder->L_name,"_T");
	bind_and_initialize(rhs,rhs_binder->L_name,"_R");
    }else{
	cs->put_string("default:{\n");
	bind_and_initialize(lhs,lhs_binder->L_name,"_T");
	bind_and_initialize(strans->type_name,rhs_binder->L_name,"_R");
    }
    cs->put_code(creator);
    cs->put_string("} break;\n");
    return;
}
void
SV_node::code_gen(CG_context c,o_stream *s)
{
    switch(c){
      case CGC_header:
	mk_class_header2(sym->name,df->LL->type_name->name,s);
	s->put_string("{\n");
	if(local_def)
	  s->put_code(local_def);
	if(transformer)
	  s->put_string("int transform(Seqn *,int);\n");
	if(flow_function || is_identity)
	  s->put_string("void flow(Snode *,Svar *);\n");
	if(process_function)
	  s->put_string("void out_action(Snode *,Svar *);\n");
	s->put_string("void _init(Snode *);\n");
	s->put_string(sym->name);
	s->put_string("(){ kind=Sk_");
	s->put_string(sym->name);
	s->put_string("; }\n};\n");
	break;
      case CGC_code:
	s->put_string("void\n");
	s->put_string(sym->name);
	s->put_string("::_init(Snode *_S){\n");
	s->put_string(" S=(");
	df->SS_pointer_type->print(s);
	s->put_string(")_S;\n}\n");
	break;
      default:
	assert(0);
    }
}
void
SV_node::gen(Symbol *s,SV_dflow *df)
{
  mk_class_header(s->name,df->LL->type_name->name);
  defs_stream->put_string("{\n");
  if(local_def)
    defs_stream->put_code(local_def);
  if(transformer)
    defs_stream->put_string("int transform(Seqn *,int);\n");
  if(flow_function || is_identity)
    defs_stream->put_string("void flow(Snode*,Svar *);\n");
  if(process_function)
    defs_stream->put_string("void out_action(Snode *,Svar *);\n");
  if(!df->new_version)
    {
      defs_stream->put_string("void _init(Snode *);\n");
      defs_stream->put_string(s->name);
      defs_stream->put_string("(){ kind=Sk_");
      defs_stream->put_string(s->name);
      defs_stream->put_string("; }\n};\n");
    }
  else
    {
      if(local_def)
	{
	  defs_stream->put_string(s->name);
	  defs_stream->put_string("(){ kind=Sk_");
	  defs_stream->put_string(s->name);
	  defs_stream->put_string("; }\n};\n");
	}
      else
	{
	  /* create a static verion */
	  defs_stream->printf("  void *operator new(size_t);\n"
			      "  %s();\n"
			      "  ~%s();\n"
			      "  void operator delete(void *);\n"
			      "};\n",
			      s->name,s->name);
	}
    }
    
  if(!df->new_version)
    {
      code_stream->printf("void %s::_init(Snode *_S)\n"
			  "{\n",
			  s->name);
      CPS(" S=(");
      df->SS_pointer_type->print(code_stream);
      CPS(")_S;\n}\n");
    }
  else
    {
      if(local_def==0){

	/*
	If a node does not have any members, then we can safely share one
	  static copy of the node.  To do this, we need to override the
	  default new operator.  Please note that the destructor tests that
	  you have not gone and created your own structure.
	  */

	code_stream->printf("static %s static_%s;\n",s->name,s->name);
	code_stream->printf("void *%s::operator new(size_t)\n"
			    "{\n"
			    "  return &static_%s;\n"
			    "}\n",
			    s->name, s->name);

	code_stream->printf("%s::%s()\n"
			    "{\n"
			    "  kind= Sk_%s;\n"
			    "}\n",
			    s->name,s->name,s->name,s->name);

	code_stream->printf("%s::~%s()\n"
			    "{\n"
			    "  assert(this==&static_%s);\n"
			    "}\n",
			    s->name, s->name, s->name);

	code_stream->printf("void %s::operator delete(void *ptr)\n"
			    "{\n"
			    "  assert(ptr==&static_%s);\n"
			    "}\n",
			    s->name, s->name);
      }
    }
}

void
SV_node::gen_func(Symbol *s,SV_dflow *df)
{
    if(transformer){
	CPS("int\n");
	CPS(s->name);
	CPS("::transform(Seqn *Q,int _MODE){\n");
	df->VV_pointer_type->print(code_stream);
	CPS(" AV = (");
	df->VV_pointer_type->print(code_stream);
	CPS(") (Q->ancestor?Q->ancestor->Q->V:0);\n");
	df->VV_pointer_type->print(code_stream);
	CPS(" V = (");
	df->VV_pointer_type->print(code_stream);
	CPS(") Q->V;\n");
	/* to get rid of warning */
//	CPS("int MODE = _MODE;\n");
	code_stream->put_code(transformer);
	CPS("}\n");
    }
    if(flow_function){
	assert(is_identity==0);
	CPS("void\n");
	CPS(s->name);
	CPS("::flow(Snode *__N,Svar *__V){\n");
	df->VV_pointer_type->print(code_stream);
	CPS(" _V = (");
	df->VV_pointer_type->print(code_stream);
	CPS(") __V;\n");
	df->SS_pointer_type->print(code_stream);
	CPS(" _N = (");
	df->SS_pointer_type->print(code_stream);
	CPS(") __N;\n");
	code_stream->put_code(flow_function);
	CPS("}\n");
    }
    if(is_identity){
	assert(flow_function==0);
	CPS("void\n");
	CPS(s->name);
	CPS("::flow(Snode *,Svar *){\n");
	CPS("}\n");
    }
    if(process_function){
	CPS("void\n");
	CPS(s->name);
	CPS("::out_action(Snode *__N,Svar *__V){\n");
	df->VV_pointer_type->print(code_stream);
	CPS(" _V = (");
	df->VV_pointer_type->print(code_stream);
	CPS(") __V;\n");
	df->SS_pointer_type->print(code_stream);
	CPS(" _N = (");
	df->SS_pointer_type->print(code_stream);
	CPS(") __N;\n");
	code_stream->put_code(process_function);
	CPS("}\n");
    }
}

/*##
frag(gensym,"Generate a unique symbol")
tex @{
Generate a unique symbol.  The assertion at the end
insists that unique is never non-zero.  And unique
can only be zero when there have been too many calls
to this routine, i.e. more than $2^{31}$ calls which
is a very unlikely event.
@}

*/
char *
SV_dflow::gensym(const char *prefix)
{
	char *n;
	if(prefix==0)
		prefix=name->name;
	n=new char[strlen(prefix)+11];
	sprintf(n,"%s%d",prefix,unique++);
	assert(unique>0);
	return n;
}

/*##
frag(mk_class,"Make a class definition")

*/
void
mk_class_header(const char *cname,const char *base)
{
	defs_stream->put_string("struct ");
	defs_stream->put_string(cname);
	if(base){
		defs_stream->put_string(": public ");
		defs_stream->put_string(base);
	}
}
void
mk_class_header2(const char *cname,const char *base,o_stream *s)
{
	s->put_string("struct ");
	s->put_string(cname);
	if(base){
		s->put_string(": public ");
		s->put_string(base);
	}
}
void
mk_class(const char *cname,const char *base,C_code *def)
{
	mk_class_header(cname,base);
	assert(def!=0);
	defs_stream->put_string("\n{");
	defs_stream->put_string(def);
	defs_stream->put_string("};\n");
}

/*##
frag(mk_name,"Make a class name")

*/
char *
mk_name(const char *name,const char *ending)
{
	char *cname=new char[strlen(name)+strlen(ending)+1];
	strcpy(cname,name);
	strcat(cname,ending);
	return cname;
}

/*##
frag(default_declare_functions,"Default @Sym_val::declare@ and @Sym_val::print@")

*/
void Sym_val::declare(o_stream *,o_stream *,const char *, sc_expr *)
	{ assert(0); }

void Sym_val::code_gen(CG_context c,o_stream *s)
{
    switch(c){
      case CGC_header:
	s->put_string("HEADER ");
	break;
      case CGC_code:
	s->put_string("CODE ");
      case CGC_decl_head:
	s->put_string("DECL_HEAD ");
	break;
      case CGC_decl_tail:
	s->put_string("DECL_TAIL ");
	break;
      case CGC_defn_head:
	s->put_string("DEFN_HEAD ");
	break;
      case CGC_defn_tail:
	s->put_string("DEFN_TAIL ");
	break;
    }
    s->put_string("unimplemented;\n");
}

void Sym_val::print(o_stream *) { assert(0); }

/*##
frag(ty_base_declare,"Base type declaration")

*/
void
ty_base::print(o_stream *os)
{
	switch(type){
	case BT_void:	os->put_string("void");	break;
	case BT_int:	os->put_string("int");	break;
	case BT_char:	os->put_string("char");	break;
	}
}
void
ty_base::code_gen(CG_context c,o_stream *s)
{
    switch(c){
      case CGC_decl_head:
	print(s);
	break;
      case CGC_decl_tail:
	break;
      default:
	assert(0);
    }
}
void
ty_base::declare(o_stream *os,o_stream *code_s,const char *n, sc_expr *e)
{
	print(os);
	os->put(' ');
	os->put_string(n);
	if(e){
		os->put('=');
		e->print(os,code_s);
	}
	os->put_string(";\n");
}

/*##
frag(ty_named_declare,"Named type declaration")

*/
void
ty_named::print(o_stream *os)
{ os->put_string(type_name->name); }
void
ty_named::code_gen(CG_context c,o_stream *s)
{
    switch(c){
      default:
	assert(0);
      case CGC_decl_head:
	print(s);
	break;
      case CGC_decl_tail:
	break;
    }
}
void
ty_named::declare(o_stream *os,o_stream *code_s,const char *n,sc_expr *e)
{
	print(os); os->put(' ');
	os->put_string(n);
	if(e){
		os->put('=');
		e->print(os,code_s);
	}
	os->put_string(";\n");
}

/*##
frag(ty_record_constructor,"@ty_record::ty_record@")

*/
ty_record::ty_record(Symbol *n,ty_record *b, C_code *o)
     :SV_type(TY_record),bases(b),members(0,SMALL_TABLE_SIZE)
{
	type_name=n; opaque=o;
}
ty_record::ty_record(Symbol *n,const ty_record_list &bl, C_code *o)
     :SV_type(TY_record),bases(bl),members(0,SMALL_TABLE_SIZE)
{
	type_name=n; opaque=o;
}

ty_record::~ty_record()
{
	if (opaque != NULL)
		delete opaque;
}

/*##
frag(ty_record_list_destructor,"@ty_record_list::~ty_record_list@")

*/
ty_record_list::~ty_record_list()
{
	while(!is_empty()) pop();
}

/*##
frag(ty_record_declare,"Record type declaration")
tex @{
This functions checks a few subtle things since this is the
first chance we get to.
If this is anonymous type, i.e. its {\tt type\_name } field is
empty then this type value better not have been derived.
It also checks to seen if the argument {\tt n} is null.
If so, then we just want to make a class declaration instead of
declaring something as having this type.
This is so that we don't need to provide a {\tt print} function
to stop class declarations from appearing inside argument list.
@}

*/
void
ty_record::code_gen(CG_context cntx,o_stream *s)
{
    ty_record_list_iter bi;
    Symbol *sym;
    Symbol_iter si;
    Type_iter ti;
    SV_type *ty;
    switch(cntx){
      case CGC_header:
	s->put_string("struct ");
	if(type_name!=0){
	    s->put_string(type_name->name);
	    if(!bases.is_empty()){
		s->put(':');
		bi.reset(&bases);
		for(;;){
		    s->put_string(" public ");
		    s->put_string(bi.step()->type_name->name);
		    if(bi.is_empty())
		      break;
		    s->put(',');
		}
	    }
	}else assert(bases.is_empty());
	s->put_string(" {\n");
	ti.reset(&type_list);
	si.reset(&member_list);
	while(!si.is_empty()){
	    assert(!ti.is_empty());
	    sym=si.step();
	    ty=ti.step();
	    TAG_ASSERT(SV_type,sym);
	    s->indent();
	    ty->code_gen(CGC_decl_head,s);
	    s->put_string(sym->name);
	    ty->code_gen(CGC_decl_tail,s);
	    s->put_string(";\n");
	}
	if(opaque)
	  s->put_string(opaque);
	s->put_string("\n} ;\n");
	break;
      case CGC_code:
	ti.reset(&type_list);
	si.reset(&member_list);
	while(!si.is_empty()){
	    assert(!ti.is_empty());
	    sym=si.step();
	    ty=ti.step();
	    TAG_ASSERT(SV_type,sym);
	    s->indent();
	    ty->code_gen(CGC_decl_head,s);
	    s->put_string(sym->name);
	    ty->code_gen(CGC_decl_tail,s);
	    s->put_string(";\n");
	}
	break;
      default:
	assert(0);
    }	
}
void
ty_record::declare(o_stream *def_s,o_stream *code_s,const char *n,sc_expr *e)
{
	ty_record_list_iter bi;
	def_s->put_string("struct ");
	if(type_name!=0){
		def_s->put_string(type_name->name);
		if(!bases.is_empty()){
			def_s->put(':');
			bi.reset(&bases);
			for(;;){
				def_s->put_string(" public ");
				def_s->put_string(bi.step()->type_name->name);
				if(bi.is_empty())
					break;
				def_s->put(',');
			}
		}
	}else assert(bases.is_empty());
	def_s->put_string(" {\n");
	for(Symbol_iter si(&member_list);!si.is_empty();){
		Symbol *s=si.step();
		TAG_ASSERT(SV_type,s);
		def_s->indent();
		((SV_type*)s->value)->declare(def_s,code_s,s->name);
	}
	if(opaque)
		def_s->put_string(opaque);
	def_s->put_string("\n} ");
	if(n) def_s->put_string(n);
	if(e){
		def_s->put('=');
		e->print(def_s,code_s);
	}
	def_s->put_string(";\n");
}

/*##
frag(ty_function_declare,"Function type declaration")

*/
void
ty_function::code_gen(CG_context cntx,o_stream *s)
{
    Type_iter ti;
    switch(cntx){
      case CGC_decl_head:
	return_type->code_gen(CGC_decl_head,s);
	s->put(' ');
	break;
      case CGC_decl_tail:
	s->put(LPAREN);
	ti.reset(&formals);
	if(!ti.is_empty())
	  for(;;){
	      SV_type *t=ti.step();
	      t->code_gen(CGC_decl_head,s);
	      if(ti.is_empty()) break;
	      s->put(',');
	  }
	s->put(RPAREN);
	return_type->code_gen(CGC_decl_tail,s);
	break;
      default:
	assert(0);
    }
}
void
ty_function::declare(o_stream *os,o_stream *code_s,const char *n,sc_expr *e)
{
	Type_iter ti(&formals);
	return_type->print(os);
	os->put(' ');
	os->put_string(n);
	os->put('(');
	if(!ti.is_empty())
		for(;;){
			SV_type *t=ti.step();
			t->print(os);
			if(ti.is_empty()) break;
			os->put(',');
		}
	os->put(')');
	if(e){
		os->put('=');
		e->print(os,code_s);
	};
	os->put_string(";\n");
}

/*##
frag(ty_pointer_declare,"Pointer type declaration")

*/
void
ty_pointer::print(o_stream *os)
{
	type->print(os);
	os->put_string(" *");
}
void
ty_pointer::code_gen(CG_context c,o_stream *s)
{
    switch(c){
      default:
	assert(0);
      case CGC_decl_head:
	switch(type->kind){
	  case TY_function:
	    type->code_gen(c,s);
	    s->put(LPAREN);
	    s->put('*');
	    break;
	  default:
	    type->code_gen(c,s);
	    s->put_string(" *");
	    break;
	}
       break;
      case CGC_decl_tail:
	switch(type->kind){
	  case TY_function:
	    s->put(RPAREN);
	    type->code_gen(c,s);
	    break;
	  default:
	    break;
	}
	break;
    }
}
void
ty_pointer::declare(o_stream *def_s,o_stream *code_s,const char *n, sc_expr *e)
{
	switch(type->kind){
	case TY_function: {
		char *nn=new char[strlen(n)+5];
		sprintf(nn,"(*%s)",n);
		type->declare(def_s,code_s,nn,e);
		delete nn;
		break;
	    }
	default:
		print(def_s);
		def_s->put_string(n);
		if(e){
			def_s->put('=');
			e->print(def_s,code_s);
		}
		def_s->put_string(";\n");
		break;
	}
}

/*##
frag(ty_enum_declare,"Enumerated type declaration")

*/
void
ty_enum::print(o_stream *os)
{
	Symbol_iter si(&member_list);
	os->put_string("enum ");
	os->put_string(type_name->name);
	os->put('{');
	if(!si.is_empty())
		for(;;){
			Symbol *s=si.step();
			os->put_string(s->name);
			if(si.is_empty()) break;
			os->put(',');
		}
	os->put('}');
}
void
ty_enum::code_gen(CG_context c,o_stream *s)
{
    Symbol_iter si;
    Symbol *sym;
    switch(c){
      default:
	assert(0);
      case CGC_decl_head:
	s->put_string("enum ");
	s->put_string(type_name->name);
	break;
      case CGC_decl_tail:
      case CGC_code:
	break;
      case CGC_header:
	code_gen(CGC_decl_head,s);
	si.reset(&member_list);
	s->put('{');
	       if(!si.is_empty())
	       for(;;){
		   sym=si.step();
		   s->put_string(sym->name);
		   if(si.is_empty())
		     break;
		   s->put(',');
	       }
	       s->put('}');
	break;
    }
}
void
ty_enum::declare(o_stream *os,o_stream *code_s,const char *n,sc_expr *e)
{
	print(os);os->put(' ');
	os->put_string(n);
	if(e){
		os->put('=');
		e->print(os,code_s);
	}
	os->put_string(";\n");
}

/*##
frag(ty_enum_add,"Add a member to an enumerated type")

*/
int
ty_enum::add_number(Symbol *n)
{
	member_list.append(n);
	return ++count;
}
int
ty_enum::add_number(char *n)
{ return add_number(new Symbol(n)); }


SV_const::SV_const(SV_type *t):Sym_val(TAG(SV_const)),type(t)
{
}

void
SV_const::declare(o_stream *,o_stream *,const char *,sc_expr *){assert(0);}
void
CN_function::declare(o_stream *def_s,o_stream *code_s,const char *n,sc_expr *)
{
	ty_function *t=(ty_function *)type;
	Type_iter ti(&t->formals);
	Symbol_iter si(&formals);
	t->return_type->print(code_s);
	code_s->put(' ');
	code_s->put_string(n);
	code_s->put('(');
	if(!ti.is_empty())
		for(;;){
			SV_type *t=ti.step();
			Symbol *s=si.step();
			t->print(code_s);
			code_s->put(' ');
			code_s->put_string(s->name);
			if(ti.is_empty()) break;
			code_s->put(',');
		}
	code_s->put_string(")\n");
	assert(si.is_empty());	// compatibility check
	body.print(def_s,code_s);
	code_s->put('\n');
}

/*##
frag(ty_record_add_member,"Add a member to a record type")

*/
void
ty_record::add_member(char *field,SV_type *type)
{
	Symbol *sp=new Symbol(field);
	sp->value=type;
	members.enter(sp);
	member_list.append(sp);
	type_list.append(type);
}

/*##
frag(default_print,"Default print for @sc_code@")

*/
void
sc_code::print(o_stream *,o_stream *)
{ assert(0); }

/*##
frag(opaque_print,"Printing @sc_opaque@ values")

*/
void
sc_opaque::print(o_stream *,o_stream *code_s)
{
	code_s->indent_level++;
	code_s->put_string("{\n");
	code_s->put_string(stuff);
	code_s->indent_level--;
	code_s->nl();
	code_s->put('}');
}

/*##
frag(print_list,"Printing @sc_list@ values")

*/
void
sc_list::print(o_stream *def_s,o_stream *code_s)
{
	code_s->indent_level++;
	code_s->put_string("{\n");
	for(Symbol_iter ni(&symbols);!ni.is_empty();){
		Symbol *n=ni.step();
		TAG_CHECK(SV_cell,n)->declare(code_s,code_s,n->name);
	}
	for(sc_iter si(this);!si.is_empty();)
		si.step()->print(def_s,code_s);
	code_s->indent_level--;
	code_s->nl();
	code_s->put_string("}");
}

/*##
frag(const_print,"Printing a constant")

*/
void
sc_const::print(o_stream *ds,o_stream *cs)
{
	constant->declare(ds,cs,0);
}
void
cn_integer::declare(o_stream *,o_stream *cs,const char *n,sc_expr *)
{
	char buf[20];
	assert(n==0);
	sprintf(buf,"%d",value);
	cs->put_string(buf);
}

/*##
frag(sc_id_print,"Printing @sc_id@ node")

*/
void
sc_id::print(o_stream *,o_stream *cs)
{
	cs->put_string(id->name);
}

/*##
frag(sc_switch_print,"Printing an @sc_switch@ node")

*/
void
sc_switch::print(o_stream *ds,o_stream *cs)
{
	sc_citer ci(&cases);
	cs->put_string("switch(");
	test->print(ds,cs);
	cs->put_string("){\n");
	while(!ci.is_empty()){
		sc_case *c=ci.step();
		sc_eiter ei(&c->labels);
		while(!ei.is_empty()){
			sc_expr *e=ei.step();
			cs->put_string("case ");
			e->print(ds,cs);
			cs->put(':');
			if(!ei.is_empty())
				cs->nl();
		}
		cs->indent_level++;
		cs->nl();
		if(c->code)
			c->code->print(ds,cs);
		cs->indent_level--;
		cs->nl();
	}
	cs->put_string("}\n");
}

/*##
frag(priority,"Precedence priorities")

*/
enum PRI { PRI_bottom, PRI_select, PRI_atom };
int sc_expr::priority() { assert(0); return 0; }
int sc_const::priority() { return PRI_atom; }
int sc_id::priority() { return PRI_atom; }
int sc_select::priority() { return PRI_select; }

/*##
frag(sc_select_print,"Printing a @sc_select@")

*/
void
sc_select::print(o_stream *ds,o_stream *cs)
{
	int need_parens=structure->priority()<priority();
	if(need_parens) cs->put('(');
	structure->print(ds,cs);
	if(need_parens) cs->put(')');
	cs->put_string("->");
	cs->put_string(field->name);
}

/*##
frag(close_def,"Close the definition of function")

*/
void
CN_function::close_def()
{
	Type_iter ti(&((ty_function*) type->check(TY_function))->formals);
	Symbol_iter si(&formals);
	while(!si.is_empty()){
		assert(!ti.is_empty());
		Symbol *s=si.step();
		SV_type *t=ti.step();
		assert(s->value==0);
		s->value=new SV_cell(t,1);
	}
	assert(ti.is_empty());
	closed++;
}

/*##
frag(type_of,"Return type of an expression")
tex @{
The virtual member function @type_of@ for the class @sc_expr@ and its
subclasses return the type of the expression.  The type of an
underived @sc_expr@ value is undefined and should never be requested.
The assertion failure provide a run-time check for this condition.
@}

*/
SV_type *sc_expr::type_of() { assert(0); return 0; }
SV_type *sc_const::type_of() { return constant->type; }
SV_type *
sc_id::type_of()
{
    assert(id->value);
    switch(id->value->kind){
      case TAG(SV_keyword):
      case TAG(SV_rule):
      case TAG(SV_node):
      case TAG(SV_dflow):
      case TAG(SV_type):
      case TAG(SV_error):
      case TAG(SV_path):
	assert(0);
      case TAG(SV_const):
	return ((SV_const *)id->value)->type;
      case TAG(SV_cell):
	return ((SV_cell *)id->value)->type;
    }
    return 0;
}

SV_type *sc_select::type_of()
{
    return TAG_CHECK(SV_type,field);
}

/*##
frag(unnamed_type,"@unnamed_type@")

*/
SV_type *SV_type::unnamed_type() { return this; }
SV_type * ty_named::unnamed_type(){ return type->unnamed_type(); }

/*##
frag(sc_select_constructor,"Construct an @sc_select@ value")

*/
sc_select::sc_select(sc_expr *r,char *f)
{
    ty_pointer *pt=(ty_pointer *)r->type_of()->check(TY_pointer);
    ty_record *rt=(ty_record *)pt->type->unnamed_type()->check(TY_record);
    field=rt->members.lookup(f);
    structure=r;
    TAG_ASSERT(SV_type,field);
}

/*##
frag(sv_cell_declare,"Declare an @sv_cell@")

*/
void
SV_cell::declare(o_stream *ds,o_stream *cs,const char *n,sc_expr *)
{
	assert(!is_arg);
	type->declare(ds,cs,n);
}
