/* file "sharlit.h" */

/*  Copyright (c) 1994 Stanford University

    All rights reserved.

    This software is provided under the terms described in
    the "suif_copyright.h" include file. */

#include <suif_copyright.h>

/* Sharlit Graphs (Definitions) */

#pragma interface

#ifndef SHARLIT
#define SHARLIT

#include <suif1/misc.h>
#include <suif1/glist.h>
#include "conslist.h"
class CFG_0;
class Snode_list;
class Snode_set;
class Snode;
class Strans;
class Svar;
class Seqn;
class Seqn_list;
class Seqn_conslist;
class Sproblem;
#define MAX_ARITY 1000

/*

An {\tt Snode\_list} is a {\em cons} lists of @Snode@s, and
the iterator {\tt Snode\_iter} provides access to the list of
{\tt Snode} elements that is independent of the actual implementation
of lists.

All three classes are derived from the cons list classes.
Most member functions are just syntactical sugaring for the general
list functions, while others, besides calling the analogous @conslist@
member function, perform some minor operations such as
updating the @count@ of @Snode@s in the list.

*/
struct Snode_e: public cons_e {
	Snode *to() { return (Snode *)data; }
	Snode_e(Snode *t):cons_e(t) {}
};
class Snode_list:public conslist {
public:
	int count;
	Snode_list() { count=0; }
	void push(Snode *s) { count++; conslist::push(s); }
	Snode *top() { return (Snode *)conslist::top(); }
	Snode *last() { return (Snode *)conslist::last(); }
	Snode *pop();
	Snode_e *f_pop() { return (Snode_e *) glist::pop(); }
	Snode *append(Snode *s)
		{ count++; conslist::append(s); return s; }
	Snode *set_append(Snode *);
	void remove(Snode *);
	int contains(Snode *);
	Snode *first() { return top(); }
	Snode_list *copy();
	void mk_empty();
	void clear() { glist::clear(); count=0; }
	void grab_from(Snode_list *l) { count=l->count;
				glist::grab_from((glist *)l); }
	void reverse(Snode_list *);
	Snode *operator[](int);
};
struct Snode_iter: public consiter {
	void reset(Snode_list *l) { consiter::reset(l); }
	Snode_iter() {}
	Snode_iter(Snode_list *l):consiter(l) {}
	Snode_e *f_step() { return (Snode_e *)glist_iter::step(); }
	Snode_e *f_peek() { return (Snode_e *)glist_iter::peek(); }
	Snode *step() { return (Snode *)consiter::step(); }
	Snode *peek() { return (Snode *)consiter::peek(); }
};

/*##
frag(sglobal_list,"Lists of variables")
tex @{
Lists of {\em variables} or @Svar@s are provided for convenience.
@}

*/
struct Svar_list: public conslist {
	Svar_list() {}
	Svar *push(Svar *v) { conslist::push(v); return v; }
	Svar *pop() { return (Svar *)conslist::pop(); }
	Svar *append(Svar *v) { conslist::append(v); return v; }
	int is_empty() { return conslist::is_empty(); }
};
struct Svar_iter: public consiter {
	void reset(Svar_list *l) { consiter::reset(l); }
	Svar_iter() {}
	Svar_iter(Svar_list *l):consiter(l) {}
	Svar *step() { return (Svar *)consiter::step(); }
	int is_empty() { return consiter::is_empty(); }
};

/*

Graph Edges --- Used in control flow analysis.

*/

class Sedge: public glist_e
{
friend class Sedge_list;

public:
  Seqn *from,*to;
  Sedge(Seqn *f,Seqn *t) { from=f; to=t; }
};

class Sedge_list:public glist
{

public:
  Sedge *push(Seqn *f,Seqn *t)
  {
    return (Sedge *)glist::push(new Sedge(f,t));
  }

  Sedge *push(Sedge *e)
  {
    return (Sedge *)glist::push(e);
  }

  void append(Seqn *f,Seqn *t)
  {
    glist::append(new Sedge(f,t));
  }

  void append(Sedge *e) 
  {
    glist::append(e);
  }

  Sedge *pop()
  {
    return (Sedge *)glist::pop();
  }

  Sedge *insert(Sedge *,Sedge *,Sedge *);

  Sedge *append_sorted(Sedge *e)
  {
    return insert(0,e,(Sedge *)head_e);
  }

  Sedge *append_sorted(Seqn *f,Seqn *t)
  {
    return append_sorted(new Sedge(f,t));
  }
};

struct Sedge_iter: public glist_iter
{
  void reset(Sedge_list *el) { glist_iter::reset(el); }
  Sedge_iter(Sedge_list *el) { reset(el); }
  Sedge_iter() {}
  Sedge *step() { return (Sedge *)glist_iter::step(); }
};

/*

A Graph Node

A Snode structure is the base class for flow graph nodes.  It has a
unique number, which is used by Sharlit to associate information with
the node.  Some older Sproblem subclasses use the fields Q, V, and L
to refer to the equation, flow variable, and flow function classes
respectively.  Newer Sproblem subclasses only use Q, not V and L.  Thus
the last two fields are essentially vestigial.

*/
class Snode
{
  public:
    int unique;
    Seqn *Q;
    Snode(Svar * = 0, Strans * = 0);
    virtual ~Snode() {}
  
};

class Svar {
public:
    signed char solved;
    Svar() { solved=0; }
    virtual ~Svar() {}
    
};

enum Strans_kinds
{
    Sk_top = -3, Sk_bottom = -2, Sk_identity = -1
  };

// This is the abstract class for flow functions
class Strans
{
  public:
    int kind;
    Strans() {}
    Strans(int k) { kind=k; }
    virtual ~Strans();
    void operator delete(void *);

    /* transform, flow, and _init are generated by sharlit */
    virtual int transform(Seqn *,int);
    virtual void flow(Snode *,Svar *);
    virtual void _init(Snode *);

    /* a utility function called during propagate
      to transform code. */
	
    virtual void in_action(Snode *,Svar *);
    virtual void out_action(Snode *,Svar *);

    virtual Strans *copy();
    virtual Strans *times_ss(Strans *);
    virtual Strans *times_su(Strans *);
    virtual Strans *star_u();
    virtual Strans *absorb_us(Strans *);
    virtual Strans *absorb_uu(Strans *);
    virtual Strans *meet_uu(Strans *);
    virtual void start_int(Sproblem *P,Seqn *Q);
    virtual void end_int(Sproblem *P,Seqn *Q);
    Strans *is(int k) { return this && k==kind ? this : 0; }
    int is_special() { return kind<Sk_identity; }
};
enum Flow_modes
{
    FM_none,
    FM_flow,
    FM_propagate
};

class Sproblem {
  protected:
    Snode *bb_head;
    void set_highpt(Seqn *,Seqn *);
    void dfs1_a(int &,int &,Seqn *,Seqn **,Sedge_list*,int);
    void dfs1(int&,int&,Seqn*,Seqn_conslist *,
	      Seqn_conslist *,Sedge_list * = 0, int = 0);
    virtual int isolve2(Seqn **,int,int = 1000000);
    virtual int isolve2(Seqn_conslist *, int = 1000000);
  public:
    CFG_0 *graph;
    int direction;
    int checking;
    int check_limit;
    int cnt, rcnt;
    int irred_cnt;
    int iteration;
    signed char debug_flags;
    signed char changed;
    signed char in_bb;
    Sproblem(CFG_0 *g,int d);
    virtual ~Sproblem();
    int start_status(Seqn *,int);
    int end_status(Seqn *,int);
    virtual void dfs(Seqn_list *,int,Seqn_conslist * = 0,
		     Seqn_conslist * =0,Sedge_list * = 0, int = 0);
    virtual Seqn *equation(Seqn *);
    virtual int isolve(Seqn **,int nq,int n = 10000);
    virtual int isolve(Seqn_conslist *l,int n);
    virtual int isolve(Snode_list *,int = 10000);
    virtual int flow_a(Seqn **,int);
    virtual int flow(Seqn_conslist *);
    virtual Snode **dependees2(Snode *,int *);
    virtual Snode **dependents2(Snode *,int *);
    virtual void printq(Seqn *,Svar * = 0,int = 0);
    virtual void debug_print(Snode *,Strans *,Svar * = 0,int = 0);
    virtual void print(char *);
    virtual int get_srcs(Svar **&,Snode *);
    virtual void clean(Seqn **,int);
    virtual Svar *new_var();
    virtual void delete_var(Svar *);
    virtual int copy_var(Svar *dst,Svar *src);
    virtual void meet(Svar *dst,Snode *s);
    
	/* access routines */
    virtual Snode *ancestor(Snode *);
};

/*##
frag(top_and_bottom,"Top and bottom transformers")

*/
extern Strans strans_top, strans_bottom, strans_identity;
extern Snode snode_nowhere;

/*##
frag(data_flow_equation,"Data flow equations and lists of them")
tex @{
An equation represents a data flow equation.
Each equation has a kind determined from it relationship to its dependents
and dependees, i.e. its role in the data flow graph.
For example, if the kind is set to @SEQN_header@ then the node is
the header for a strongly connected component, while @SEQN_dominator@ would
mean the node is the header of a dominator strong component.  The latter
is used in the localization of irreducibilities in the flow graph.
A kind value of @SEQN_noncyclic@ refer to the start of a non-cyclic
region; that is, it is not strongly connected but should be treated much
like one.  The only difference is that in @Sproblem::flow@ the header for
a strongly connected region is evaluated {\em after} all the nodes
in the region while the header of a non-cyclic region is evaluated first.
This can reduce the number of iterations required in @Sproblem::isolve@.

A kind value of @SEQN_bb@ is a basic block.  It is the head of
a straight line sequence of nodes.  It is treated in exactly
the same manner as an @SEQN_noncyclic@ region except there
is no nesting.
@}

*/
enum SEQN_type { SEQN_normal, SEQN_noncyclic, SEQN_bb,
		 SEQN_header, SEQN_dominator,
		 SEQN_improper, SEQN_unreachable };
class Seqn: public glist_e {
  public:
    signed char needs_var;
    signed char kind;
    unsigned short rcount, number, snumber;
    Snode *highpt;
    Snode *ancestor;
    Seqn *covered;
    Strans *iL;
    Svar *V;
    Strans *L;
    Snode *S;
    int HIGHPT() { return highpt ? highpt->Q->number : -1; }
    int solved() { return V && V->solved; }
    int depth();
    void init();
    void init(Svar *v,Strans *l)
      {
	V=v;
	L=l;
    }
    void _init(Snode *s)
      {
	  /* initialize for constructor */
	  glist_e::next_e=0;
	  ancestor=0; V=0; L=0; S=s;
	  needs_var=0; rcount=0;
	  number=snumber=0;
	  highpt=0;
	  kind=SEQN_normal;
	  iL=0;
	  covered=0;
      }
    Seqn() {}
    Seqn(Snode *s)
      {
	  _init(s);
      }
    Seqn(Svar *g,Strans *l,Snode *s)
      {
	  ancestor=0; init(g,l); S=s;
	  needs_var=0; rcount=0;
	  number=snumber=0;
	  highpt=0;
	  kind=SEQN_normal;
	  iL=0;
	  covered=0;
      }
    Seqn(Seqn *q)
      {
	  ancestor=q->ancestor; iL=q->iL; V=q->V; L=q->L; S=q->S;
	  needs_var=q->needs_var; rcount=q->rcount;
	  number=q->number;
	  snumber=q->snumber;
	  highpt=q->highpt;
	  kind=q->kind;
	  covered=0;
      }
    void operator=(Seqn &q)
      {
	  ancestor=q.ancestor; iL=q.iL; V=q.V; L=q.L; S=q.S;
	  needs_var=q.needs_var; rcount=q.rcount;
	  number=q.number;
	  snumber=q.snumber;
	  highpt=q.highpt;
	  kind=q.kind;
	  covered=0;
      }
    Seqn *pop();
    void push(Seqn *);
    Seqn *top();
};
struct Seqn_list: public glist {
	Seqn_list() {}
	~Seqn_list();
	void push(Seqn *e) { glist::push(e); }
	void append(Seqn *e) { glist::append(e); }
	Seqn *pop() { return (Seqn *)glist::pop(); }
	Seqn *first() { return (Seqn *)head_e; }
	void init();
};
struct Seqn_iter: public glist_iter {
	void reset (Seqn_list *l){glist_iter::reset(l); }
	Seqn_iter() {}
	Seqn_iter(Seqn_list *l) { reset(l); }
	Seqn *step() { return (Seqn *)glist_iter::step(); }
	int is_empty() { return glist_iter::is_empty(); }
};

/*##
frag(seqn_conslist,"Cons list for @Seqn@")

*/
struct Seqn_conslist: public conslist {
	Seqn_conslist() {}
	void push(Seqn *e) { conslist::push(e); }
	void append(Seqn *e) { conslist::append(e); }
	Seqn *pop() { return (Seqn *)conslist::pop(); }
	void f_push(cons_e *e) { glist::push(e); }
	void f_append(cons_e *e) { glist::append(e); }
	cons_e *f_pop() { return (cons_e *)glist::pop(); }
	Seqn *top() { return (Seqn *)((cons_e *)head_e)->data; }
	void init();
};
struct Seqn_consiter: public consiter {
	void reset (Seqn_conslist *l){ consiter::reset(l); }
	Seqn_consiter() {}
	Seqn_consiter(Seqn_conslist *l) { reset(l); }
	Seqn *step() { return (Seqn *)consiter::step(); }
	Seqn *peek() { return (Seqn *)consiter::peek(); }
};

class Sprob_tarjan: public Sproblem {
  protected:
    virtual Strans *eval(Snode *,Snode *);
    virtual Strans *eval(Snode *);
    virtual void path_compress(Snode *);
    virtual void R(Seqn_consiter *);
    virtual int propagate(Seqn_conslist *l);
    virtual int F(Seqn_consiter *);
    virtual int nonterminal(Strans *);
    virtual void keep(Snode *);
    virtual void link(Snode *h,Snode *q);
    virtual void update(Snode *q,Strans *P,Strans *Q);
    virtual void RO_a(int &,Seqn **,int &,Seqn **,int);
  public:
    virtual void reduction_order_a(Seqn **,Seqn **,int,	Sedge_list *);
    virtual int reducer(Seqn_conslist *,Seqn_conslist *);
    virtual int isolve(Seqn_conslist *l,int n);
    virtual int isolve(Snode_list *,int = 10000);
    Sprob_tarjan(CFG_0 *g,int d);
};
/* miscellaneous functions for Sprob_tarjan */
int is_cycle(Snode *,Snode *);

/*##
frag(Sprob_tarjan_debug,"class @Sprob_tarjan_DEBUG@")

*/
class Sprob_tarjan_DEBUG: public Sprob_tarjan {
protected:
	int debug_reduce;
	virtual int F(Seqn_consiter *);
	virtual int flow(Seqn_conslist *);
//	virtual void R(Seqn_consiter *);
public:
	Sprob_tarjan_DEBUG(CFG_0 *g,int d);
};

/*##
frag(sproblem_debug,"class @Sproblem_DEBUG@")

*/
class Sproblem_DEBUG: public Sproblem {
public:
	Sproblem_DEBUG(CFG_0 *g,int d);
	virtual int flow(Seqn_conslist *);
};
#endif
